import numpy as np
import jax.numpy as jnp
from jax import vmap,grad,jit
from jax.tree_util import Partial
import optax

from .lddmm import LDDMMLoss
from .loss import VarifoldLoss
from .optimizer import BatchIteratedBarycenterOptimizer



####################################################################################################################################
####################################################################################################################################
### GENERAL ###
####################################################################################################################################
####################################################################################################################################

def _argmedian_length_from_mask(x):
    return np.argpartition(x, np.sum(x) // 2)[len(x) // 2]

def BarycenterLDDMMLoss(Kv,dataloss:callable,gamma_loss =0.,nt=10,deltat=1.0): 
    """ 
    Return the Barycenter LDDDMMLoss function related to the barycenter problem

    Parameters
    ----------
        Kv : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        dataloss: (X,X_mask,Y,Y_mask)->float
            the attachment loss function, typically a varifold loss
        gamma: float 
            regularization constant related to the norm of the initial velocity
        nt: int
            number of integration time points, default=10
        delta : float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt, default to 1
            

    Returns
    -------
        LDDMM registration loss : (P,X,X_mask,Y,Y_mask)->float
         :math:`\\sum_i \\gamma |v_i|_{\\mathsf{V}}+\\mathcal{L}(\\phi^{v_i}\\cdot X,Y_i)` where X is an array of size (n_samples,d+1),
        and P are the momentums (n_samples,d+1) such that :math:`\\phi^{v_0}=\\exp_X(P)` is the exponential map which is computed by integrating the Hamiltonian system
        using the Ralston integrator on nt points. Also called shooting function.
    """
    unit_loss = LDDMMLoss(Kv,dataloss,gamma_loss,nt,deltat)
    f = vmap(unit_loss,(0,None,None,0,0),0)
    def bloss(p0,q0,q0_mask,q1,q1_mask): 
        return jnp.sum(f(p0,q0,q0_mask,q1,q1_mask))
    return bloss


# voir si on fait sauter time initializer
def batch_barycenter_registration(shoot,batched_q1,batched_q1_mask,Kv,dataloss:callable,init=None,init_mask=None,time_initializer=None,niter=100, optimizer = optax.adam(learning_rate=0.1),gamma_loss =0.,nt=10,deltat=1.0,update_interval=100,verbose=True,stream_bool=False,stream_object=None):
    """ 
    Return the Barycenter LDDDMMLoss function related to the barycenter problem

    Parameters
    ----------
        shoot : (p0,q0,q0_mask)->(p1,q1)
            Shooting function, return the Hamiltonian dynamics at time 1 starting from (p0,q0)
        batched_q1 : array of shape (N_timeseries,batch_size,n_samples,d+1)
            target time series' graphs, in practice batch_size is set to 1
        batched_q1_mask : array of shape (N_timeseries,batch_size,n_samples,1)
            target time series' graphs mask
        Kv : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        dataloss: (X,X_mask,Y,Y_mask)->float
            the attachment loss function, typically a varifold loss
        init : array of shape (N_timeseries,batch_size,n_samples,d+1)
            initial momentums for the optimizers, default None
        init_mask : array of shape (N_timeseries,batch_size,n_samples,1)
            initial momentums mask for the optimizers, default None
        niter : int
            number of iterations for the optimizer
        optimizer : optimize object from optax
            Default to optax.adabelief(learning_rate=0.1)
        gamma_loss: float 
            regularization constant related to the norm of the initial velocity
        nt: int
            number of integration time points, default=10
        delta : float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt, default to 1
        update_interval : int
            The barycenter is updated every update_interval steps by shooting the current barycental using the average of the estimated momentums,default 100
            

    Returns
    -------
        batched_p : array of shape (n_samples,d+1)
            momentums minimizing the LDDMM loss related to the multiple shooting problems
        q0 : array of shape (n_samples,d+1)
            learned barycenter = time series' graph of referece
        q0_mask : array of shape (n_samples,1)
            learned barycenter time series' graph mask
    """
    if init is None: 
        sigs_size = np.sum(batched_q1_mask,axis=2)
        median_length_idx = np.unravel_index(np.argsort(sigs_size.flatten(),kind='stable')[sigs_size.shape[0]//2],sigs_size.shape)[:-1]
        q0 = batched_q1[median_length_idx]
        q0_mask = batched_q1_mask[median_length_idx]
    else: 
        q0 = init
        q0_mask = init_mask
    if time_initializer is None:
        batched_p0 = jnp.zeros_like(batched_q1,dtype = jnp.float32)
    if callable(time_initializer): 
        if verbose: 
            print("Time initialization")
        batched_p0,q0,q0_mask = time_initializer(batched_q1,batched_q1_mask,q0,q0_mask)
    bloss = BarycenterLDDMMLoss(Kv,dataloss,gamma_loss,nt,deltat)
    opt = BatchIteratedBarycenterOptimizer(shoot,bloss,niter,optimizer, update_interval =update_interval,verbose=verbose,stream_bool=stream_bool,stream_object=stream_object)
    return *opt(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask),q0_mask

def batch_varifold_barycenter_registration(batched_q1,batched_q1_mask,Kv,Kl,init=None,init_mask=None,time_initializer=None,niter=100, optimizer = optax.adam(learning_rate=0.1),gamma_loss =0.,nt=10,deltat=1.0,verbose=True):
    dataloss = VarifoldLoss(Kl)
    return batch_barycenter_registration(batched_q1,batched_q1_mask,Kv,dataloss,init,init_mask,time_initializer,niter,optimizer,gamma_loss,nt,deltat,verbose)



import numpy as np
import jax.numpy as jnp

#making more modular using a triangular structure + a kernel ? Not necessary


def Varifold_TSLDDMM_Gaussian_Kernel(t_sigma_1:float,s_sigma_1:float,t_sigma_2:float,s_sigma_2:float,np_dtype = np.float32):
    """
    Generate TS-LDDMM  Varifold kernel function,  using the notations of the paper "Shape analysis for time series"
    :math:`k((x,v),(x',v'))=k_{\\text{pos}}(x,x')k_{\\text{dir}}(v,v')` where
    :math:`k_{\\text{pos}}((t,x),(t',x'))=K^{(1)}_{\\sigma_{\\text{pos},t}}(t,t')K^{(d)}_{\\sigma_{\\text{pos},x}}(x,x')`, 
    :math:`k_{\\text{dir}}((t,x),(t',x'))=K^{(1)}_{\\sigma_{\\text{dir},t}}(t,t')K^{(d)}_{\\sigma_{\\text{dir},x}}(x,x')`
    and 
    :math:`K_{\\sigma}^{(a)}(w,w')=\\exp(-|w-w'|^2/\\sigma))` the :math:`a` dimensional isotropic Gaussian kernel of variance :math:`\\sigma`.

    Parameters
    ----------
        t_sigma_1 : (float)
             variance of the kernel related to time coordinate of the kernel related to position :math:`\\sigma_{\\text{pos},t}`

        s_sigma_1 : (float)
             variance of the kernel related to space coordinate of the kernel related to position :math:`\\sigma_{\\text{pos},x}`

        t_sigma_2 : (float)
             variance of the kernel related to time coordinate of the kernel related to direction :math:`\\sigma_{\\text{dir},t}`

        s_sigma_2 : (float)
             variance of the kernel related to space coordinate of the kernel related to direction :math:`\\sigma_{\\text{dir},x}`

        np_dtype : (_type_, optional):
             Defaults to np.float32.

    Returns
    -------
        kernel function : (X,mask_X,Y,mask_Y,b)-> array of the shape of b
            :math:`K(X,Y)b` where X and Y are array of size (n_samples,d+1), :math:`K(X,Y)` is the kernel matrix :math:`(k(x_i,y_j))`
            and b is an array of shape (n_samples,d) with d the dimension of the problem

    """
    # important for the varifold loss 
    t_oos_1, s_oos_1  = 1/t_sigma_1, 1/s_sigma_1
    t_oos_2, s_oos_2 = 1/t_sigma_2, 1/s_sigma_2
    def K(x,y,u,v,mask_xu,mask_yv,b): 
        n_d = x.shape[1]
        oos_1 = np.ones(n_d,dtype=np_dtype).reshape(1,-1)
        oos_1[0,0], oos_1[0,1:] = t_oos_1, s_oos_1
        k1 = jnp.exp(-jnp.sum(((x*oos_1)[:,None,:]-(y*oos_1)[None,:,:])**2,axis=2))
        oos_2 = np.ones(n_d,dtype=np_dtype).reshape(1,-1)
        oos_2[0,0], oos_2[0,1:] = t_oos_2,s_oos_2
        k2 = jnp.exp(-jnp.sum(((u*oos_2)[:,None,:]-(v*oos_2)[None,:,:])**2,axis=2))
        mask = (mask_xu*mask_yv.T)
        return (k1*k2*mask)@b
    return K
def Velocity_TSLDDMM_Gaussian_Kernel(c_0:float,c_1:float,t_sigma_1:float,t_sigma_2:float,s_sigma:float): 
    """
    Function related to the velocity TSLDDMM gaussian kernel defined as,  using the notations of the paper "Shape analysis for time series" :math:`K_G((t,x),(t',x'))=\\begin{pmatrix} c_0 K_{\\text{time}}(t,t') & 0 \\\\ 0& c_1 K_{\\text{space}}((t,x),(t',x'))\\end{pmatrix}`
    with :math:`K_{\\text{time}}=K_{\\sigma_{T,0}}^{(1)}(t,t')`, :math:`K_{\\text{space}}=K_{\\sigma_{T,1}}^{(1)}(t,t')K_{\\sigma_{x}}^{(d)}(x,x')`
    and :math:`K_{\\sigma}^{(a)}(w,w')=\\exp(-|w-w'|^2/\\sigma))` the :math:`a` dimensional isotropic Gaussian kernel of variance :math:`\\sigma`.

    Parameters
    ----------

        c_0 : (float)
            time scaling :math:`c_0`

        c_1 : (float)
            space scaling :math:`c_1`

        t_sigma_1 : (float)
            the time variance parameter 1 :math:`\\sigma_{T,0}`

        t_sigma_2 : (float)
            the time variance parameter 2 :math:`\\sigma_{T,1}`

        s_sigma : (float)
            the space variance parameter :math:`\\sigma_{x}`

    Returns
    -------

        kernel function : (X,mask_X,Y,mask_Y,b)-> array of the shape of b
            :math:`K_G(X,Y)b` where X and Y are array of size (n_samples,d+1), :math:`K(X,Y)` is the kernel matrix :math:`(k(x_i,y_j))`
            and b is an array of shape (d,) with d the dimension of the problem

    """
    #variance parameter
    t_oos_12 = 1/t_sigma_1**2
    t_oos_22 = 1/t_sigma_2**2
    s_oos2 = 1/s_sigma**2
    def K(x,mask_x,y,mask_y,b): 
        #return K_X,Y@b
        t_x, s_x = x[:,:1], x[:,1:]
        t_y, s_y = y[:,:1], y[:,1:]
        t_b, s_b = b[:,:1], b[:,1:]
        #time difference, for the kernel related to time
        time_sum = jnp.sum((t_x[:,None,:]-t_y[None,:,:])**2,axis=2)
        t_res_1 = jnp.exp(-t_oos_12*time_sum)
        # value difference, for the kernel related to space
        s_res = jnp.exp(-s_oos2*jnp.sum((s_x[:,None,:]-s_y[None,:,:])**2,axis=2)-t_oos_22*time_sum)
        mask = mask_x*mask_y.T
        #aggregation (mu*K_time@t_b,lambda*K_space@s_b)
        return jnp.hstack((c_0*(mask*t_res_1)@(t_b*mask_y), c_1*(mask*s_res)@(s_b*mask_y)))
    return K 



def Velocity_TSLDDMM_Cauchy_Kernel(c_0:float,c_1:float,t_sigma_1:float,t_sigma_2:float,s_sigma:float): 
    """
    Function related to the velocity TSLDDMM cauchy kernel defined as, using the notations of the paper "Shape analysis for time series"
    :math:`K_G((t,x),(t',x'))=\\begin{pmatrix} c_0 K_{\\text{time}}(t,t') & 0 \\\\ 0& c_1 K_{\\text{space}}((t,x),(t',x'))\\end{pmatrix}`
    with :math:`K_{\\text{time}}=K_{\\sigma_{T,0}}^{(1)}(t,t')`,  :math:`K_{\\text{space}}=K_{\\sigma_{T,1}}^{(1)}(t,t')K_{\\sigma_{x}}^{(d)}(x,x')`
    and :math:`K_{\\sigma}^{(a)}(w,w')=\\frac{1}{1+|w-w'|^2/\\sigma)}` the :math:`a` dimensional isotropic Cauchy kernel of variance :math:`\\sigma`.

    Parameters
    ----------
        c_0 : (float)
            time scaling :math:`c_0`

        c_1 : (float)
            space scaling :math:`c_1`

        t_sigma_1 : (float)
            the time variance parameter 1 :math:`\\sigma_{T,0}`

        t_sigma_2 : (float)
            the time variance parameter 2 :math:`\\sigma_{T,1}`

        s_sigma : (float)
            the space variance parameter :math:`\\sigma_{x}`

    Returns
    -------
    
        kernel function : (X,mask_X,Y,mask_Y,b)-> array of the shape of b
            :math:`K_G(X,Y)b` where X and Y are array of size (n_samples,d+1), :math:`K(X,Y)` is the kernel matrix :math:`(k(x_i,y_j))`
            and b is an array of shape (d,) with d the dimension of the problem


    """
    t_oos_12 = 1/t_sigma_1**2
    t_oos_22 = 1/t_sigma_2**2
    s_oos2 = 1/s_sigma**2
    def K(x,mask_x,y,mask_y,b): 
        t_x, s_x = x[:,:1], x[:,1:]
        t_y, s_y = y[:,:1], y[:,1:]
        t_b, s_b = b[:,:1], b[:,1:]
        time_sum = jnp.sum((t_x[:,None,:]-t_y[None,:,:])**2,axis=2)
        t_res_1 = 1/(1+t_oos_12*time_sum)
        s_res = 1/(1+s_oos2*jnp.sum((s_x[:,None,:]-s_y[None,:,:])**2,axis=2)+t_oos_22*time_sum)
        mask = mask_x*mask_y.T
        return jnp.hstack((c_0*(mask*t_res_1)@(t_b*mask_y), c_1*(mask*s_res)@(s_b*mask_y)))
    return K 








import numpy as np
import matplotlib.pyplot as plt

#see if we use it

def plot2Dfigure(s_sig,t_sig,p0,shoot,flow,mask_s_sig=None,mask_t_sig=None,nv=50,nh=50): 
    if mask_s_sig is None:
        s_sig = (s_sig,)
        plot_s_sig = s_sig 
    else: 
        plot_s_sig = s_sig[np.where(mask_s_sig==True)[0],:]
        s_sig = (s_sig,mask_s_sig)
  
    if mask_t_sig is None: 
        t_sig = (t_sig,)
        plot_t_sig = t_sig
    else: 
        plot_t_sig = t_sig[np.where(mask_t_sig==True)[0],:]
        t_sig = (t_sig,mask_t_sig)
        

    _,p_sig = shoot(p0,*s_sig)
    if mask_s_sig is None: 
        plot_p_sig = p_sig
    else: 
        plot_p_sig = p_sig[np.where(mask_s_sig==True)[0],:]

    fig,axs = plt.subplots(2,1,sharex=True,sharey=True, figsize = (10,5))
    axs[0].plot(*plot_s_sig.T, label = "from")
    axs[0].legend()
    axs[1].plot(*plot_t_sig.T, color = "green", label = "to")
    axs[1].plot(*plot_p_sig.T, color = "r", label = "warping")
    axs[1].legend()
    sz = 0.2 
    a = min(np.min(plot_s_sig[:,0]),np.min(plot_t_sig[:,0]),np.min(plot_p_sig[:,0]))
    b = max(np.max(plot_s_sig[:,0]),np.max(plot_t_sig[:,0]),np.max(plot_p_sig[:,0]))
    c = min(np.min(plot_s_sig[:,1]),np.min(plot_t_sig[:,1]),np.min(plot_p_sig[:,1]))
    d = max(np.max(plot_s_sig[:,1]),np.max(plot_t_sig[:,1]),np.max(plot_p_sig[:,1]))
    lsp1 = np.linspace(a-sz*(b-a),b+sz*(b-a),nv,dtype=np.float32)
    lsp2 = np.linspace(c-sz*(d-c),d+sz*(d-c),nh,dtype=np.float32)
    X1, X2 = np.meshgrid(lsp1,lsp2)
    axs[0].plot(X1,X2,'k',linewidth=.25)
    axs[0].plot(X1.T,X2.T,'k',linewidth=.25)
    x = np.concatenate((X1.reshape(nv*nh,1),X2.reshape(nv*nh,1)),axis=1)
    phix = flow(x,p0,*s_sig)[0]
    X1 = phix[:,0].reshape(nh,nv)
    X2 = phix[:,1].reshape(nh,nv)
    axs[1].plot(X1,X2,'k',linewidth=.25)
    axs[1].plot(X1.T,X2.T,'k',linewidth=.25)
    return fig,axs


import numpy as np 
import jax.numpy as jnp
import jax.lax as lax
from jax import jit, grad, vmap, jacfwd
import optax
from .loss import VarifoldLoss
from .optimizer import RegistrationOptimizer, BatchOneToManyRegistrationOptimizer

# Comprendre le code, qu'est ce qui est TS-LDDMM est ce nécessaire d'être plus modulaire ?
# Notamment si l'on ne veut faire qu'une déformation temporel ?

# TO have only a deformation on time, we can consider a Kernel that (X,mask_X,Y,mask_Y,b)-> array of shape (n_samples,1) (1 to consider only time coordinate)
# The shape of momentum should be adapted 
# no problem as long as there is not matrix inversion


####################################################################################################################################
####################################################################################################################################
### GENERAL ###
####################################################################################################################################
####################################################################################################################################

def Hamiltonian(K:callable): 
    """ 
    Return the Hamiltonian function related to the kernel K

    Parameters
    ----------

        Kernel function : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
            

    Returns
    -------

        Hamiltonian function : (P,X,mask_X)-> float
        :math:`P^\\top K(X,X)P/2` where X is an array of size (n_samples,d+1), and :math:`K(X,X)` is the kernel matrix :math:`(k(x_i,x_j))`
        and P are the momentums, it's an array of size (n_samples,d+1)
    """
    def H(p,q,q_mask): 
        # p^\top K_q,q @p/2
       return 0.5*jnp.sum(p*K(q,q_mask,q,q_mask,p))
    return H 

def HamiltonianSystem(K:callable):
    """ 
    Return the Hamiltonian system function related to the kernel K

    Parameters
    ----------

        Kernel function : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
            

    Returns
    -------

        Hamiltonian system function : (P,X,mask_X)-> (-grad_H_pos,grad_H_momen)
        :math:`(-\\nabla_X H(X,P),\\nabla_P H(X,P))` where X is an array of size (n_samples,d+1), and :math:`K(X,X)` is the kernel matrix :math:`(k(x_i,x_j))`
        and P are the momentums, it's an array of size (n_samples,d+1)
    """
    H = Hamiltonian(K)
    def HS(p,q,q_mask):
        Gp,Gq = grad(H,(0,1))(p,q,q_mask)
        return -Gq,Gp
    return HS

def Shooting(K:callable,nt=10,deltat=1.):
    """ 
    Return the Hamiltonian system function related to the kernel K

    Parameters
    ----------

        Kernel function : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        nt: int
            number of integration time points
        deltat: float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt
            

    Returns
    -------

        Exponential flow function : (P,X,mask_X)-> (Exp_X(deltat P),Exp_X'(deltat P))
        :math:`(\\exp_X(\\delta P),\\exp_X'(\\delta P))` where X is an array of size (n_samples,d+1),
        and P are the momentums, it's an array of size (n_samples,d+1). The exponential map is computed by integrating the Hamiltonian system
        using the Ralston integrator. Also called shooting function.
    """
    
    HSystem = HamiltonianSystem(K)
    dt = deltat/nt
    def body_function(i,arg): 
        p,q,q_mask = arg
        dp,dq = HSystem(p,q,q_mask)
        pi,qi = p+(2*dt/3)*dp,q+(2*dt/3)*dq
        dpi,dqi = HSystem(pi,qi,q_mask)
        p,q = p+0.25*dt*(dp+3*dpi),q+0.25*dt*(dq+3*dqi)
        return p,q,q_mask
    def f(p,q0,q0_mask):
        p,q0,q0_mask = lax.fori_loop(0,nt,body_function,(p,q0,q0_mask))
        return p,q0
    return f



def LDDMMLoss(K:callable,dataloss:callable,gamma =0.001,nt=10,deltat=1.0): 
    """ 
    Return the Loss function related to a registration problem

    Parameters
    ----------

        Kernel function : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        dataloss: (X,X_mask,Y,Y_mask)->float
            the attachment loss function, typically a varifold loss
        gamma: float 
            regularization constant related to the norm of the initial velocity
        nt: int
            number of integration time points, default=10
        delta : float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt, default to 1
            

    Returns
    -------

        LDDMM registration loss : (P,X,X_mask,Y,Y_mask)->float
        :math:`\\gamma |v_0|_{\\mathsf{V}}+\\mathcal{L}(\\phi^{v_0}\\cdot X,Y)` where X is an array of size (n_samples,d+1),
        and P are the momentums, it's an array of size (n_samples,d+1). :math:`\\phi^{v_0}=\\exp_X(P)` is the exponential map which is computed by integrating the Hamiltonian system
        using the Ralston integrator on nt points. Also called shooting function.
    """
    Hm = Hamiltonian(K)
    shoot =Shooting(K,nt,deltat)
    def loss(p0,q0,q0_mask,q1,q1_mask): 
        p,q = shoot(p0,q0,q0_mask)
        return gamma * Hm(p0,q0,q0_mask) + dataloss(q,q0_mask,q1,q1_mask)
    return loss

def Flowing(K:callable,nt=10,deltat=1.): 
    """ 
    Return the Exponential map function related to the kernel K that map as well a grid

    Parameters
    ----------

        Kernel function : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        nt: int
            number of integration time points
        deltat: float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt
            

    Returns
    -------

        Exponential flow function, taking a grid as well in input : (Grid,P,X,mask_X)-> (Grid_shooted,Exp_X(deltat P),Exp_X'(deltat P))
        :math:`(Grid_shooted,\\exp_X(\\delta P),\\exp_X'(\\delta P))` where X is an array of size (n_samples,d+1),
        and P are the momentums, it's an array of size (n_samples,d+1),
        Grid is another array that may be different than X, Grid_shooted is its final position after flowing by using the integrated velocity. The exponential map is computed by integrating the Hamiltonian system
        using the Ralston integrator. Also called shooting function.
    """
    
    HSystem = HamiltonianSystem(K)
    def FSystem(x,p,q,q_mask): 
        x_mask = jnp.full_like(x[:,:1],True,dtype=np.bool_)
        return (K(x,x_mask,q,q_mask,p),) + HSystem(p,q,q_mask) # On concatene les tuples pour rajouter un moment adapté au x, c'est pour la grille
    dt = deltat/nt
    def body_function(i,arg): 
        x,p,q,q_mask = arg
        dx,dp,dq = FSystem(x,p,q,q_mask)
        xi,pi,qi = x+(2*dt/3)*dx,p+(2*dt/3)*dp,q+(2*dt/3)*dq
        dxi,dpi,dqi = FSystem(xi,pi,qi,q_mask)
        x,p,q = x+0.25*dt*(dx+3*dxi),p+0.25*dt*(dp+3*dpi),q+0.25*dt*(dq+3*dqi)
        return x,p,q,q_mask
    def f(x0,p0,q0,q0_mask):
        x,p,q,q_mask = lax.fori_loop(0,nt,body_function,(x0,p0,q0,q0_mask))
        return x,p,q 
    return f



def DeformationGradient(K:callable,nt=10,deltat=1.0): 
    """ 
    Return the gradient of the exponential map according to the starting points

    Parameters
    ----------

        Kernel function : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        nt: int
            number of integration time points, default=10
        delta : float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt, default to 1
            

    Returns
    -------

        Gradient : (P,X,X_mask)->Jacobian_position(Shooting(K,nt,deltat))
        :math:`\\nabla_X \\exp_X(P)` where X is an array of size (n_samples,d+1),
        and P are the momentums, it's an array of size (n_samples,d+1). :math:`\\exp_X` is the exponential map which is computed by integrating the Hamiltonian system
        using the Ralston integrator on nt points. Also called shooting function.

    """
    jacHri = jacfwd(Shooting(K,nt,deltat),1)
    def gradient(p0,q0,q0_mask): 
        gr = jacHri(p0,q0,q0_mask)[1]
        idx = np.arange(gr.shape[0])
        gr = gr[idx,:,idx,:]-1
        return gr
    return jit(gradient)

#stream_bool
#stream_object for streamlit
#optax is the best 


def registration(q0,q0_mask,q1,q1_mask,Kv,dataloss,p0 = None,gamma_loss=0.0,niter=100,optimizer = optax.adabelief(learning_rate=0.1),nt=10,deltat=1.0,verbose=True,stream_bool=False,stream_object=None):
    """ 
    Return the momentum p such that the hamiltonian dynamics starting from (q_0,p) reach approximately (q_1,p_1) at time t=1, the shooting problem is performed by minimizing the LDDMM loss defined previously

    Parameters
    ----------

        q0 : array of shape (n_samples,d+1)
            initial time series' graph
        q0_mask : array of shape (n_samples,1)
            initial time series' graph mask
        q1 : array of shape (n_samples,d+1)
            target time series' graph
        q1_mask : array of shape (n_samples,1)
            target time series' graph mask
        Kv : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        dataloss: (X,X_mask,Y,Y_mask)->float
            the attachment loss function, typically a varifold loss
        p0 : array of shape (n_samples,d+1)
            initial time series' graph momentum (facultative)
        gamma_loss: float 
            regularization constant related to the norm of the initial velocity
        niter : int
            number of iterations for the optimizer
        optimizer : optimize object from optax
            Default to optax.adabelief(learning_rate=0.1)
        nt : int
            number of integration time points, default=10
        delta : float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt, default to 1
        verbose : Boolean
            Default to True, print loss every 10 iterations
        
            

    Returns 
    -------

        p : array of shape (n_samples,d+1)
             initial momentum minimizing the LDDMM loss related to the shooting problem
        q0 : array of shape (n_samples,d+1)
            initial time series' graph
        q0_mask : array of shape (n_samples,1)
            initial time series' graph mask
    """
    loss = LDDMMLoss(Kv,dataloss,gamma_loss,nt,deltat)
    opt = RegistrationOptimizer(loss,niter,optimizer,verbose=verbose,stream_bool=stream_bool,stream_object=stream_object)
    if p0 is None:
        p0 = jnp.zeros_like(q0)
    if callable(p0):
        if verbose:
            print("Time initialization")
        p0 = p0(q0,q0_mask,q1,q1_mask)
    p = opt(p0,q0,q0_mask,q1,q1_mask)
    return p,q0,q0_mask



#The following function should be used in a class in a second time

def varifold_registration(q0,q0_mask,q1,q1_mask,Kv,Kl,p0=None,gamma_loss=0.0,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=True,stream_bool=False,stream_object=None): 
    """
    Return the momentum p such that the hamiltonian dynamics starting from (q_0,p) reach approximately (q_1,p_1) at time t=1, the shooting problem is performed by minimizing the LDDMM loss defined previously

    Parameters
    ----------

        q0 : array of shape (n_samples,d+1)
            initial time series' graph
        q0_mask : array of shape (n_samples,1)
            initial time series' graph mask
        q1 : array of shape (n_samples,d+1)
            target time series' graph
        q1_mask : array of shape (n_samples,1)
            target time series' graph mask
        Kv : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        Kl : (X,mask_X,Y,mask_Y,b)-> array of the shape of b
            The Varifold kernel function. :math:`K(X,Y)b` where X and Y are array of size (n_samples,d+1), :math:`K(X,Y)` is the kernel matrix :math:`(k(x_i,y_j))`
            and b is an array of shape (n_samples,d) with d the dimension of the problem
        p0 : array of shape (n_samples,d+1)
            initial time series' graph momentum (facultative)
        gamma_loss: float 
            regularization constant related to the norm of the initial velocity
        niter : int
            number of iterations for the optimizer
        optimizer : optimize object from optax
            Default to optax.adabelief(learning_rate=0.1)
        nt : int
            number of integration time points, default=10
        delta : float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt, default to 1
        verbose : Boolean
            Default to True, print loss every 10 iterations
        
    Returns 
    -------

        p : array of shape (n_samples,d+1)
             initial momentum minimizing the LDDMM loss related to the shooting problem using the varifold loss as data attachment term
        q0 : array of shape (n_samples,d+1)
            initial time series' graph
        q0_mask : array of shape (n_samples,1)
            initial time series' graph mask
    """
    dataloss = VarifoldLoss(Kl)
    return registration(q0,q0_mask,q1,q1_mask,Kv,dataloss,p0,gamma_loss,niter,optimizer,nt,deltat,verbose=verbose,stream_bool=stream_bool,stream_object=stream_object)


def batch_one_to_many_registration(q0,q0_mask,batched_q1,batched_q1_mask,Kv,dataloss,batched_p0=None,gamma_loss=0.0,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=True,stream_bool=True,stream_object=None):
    """
    Return the momentums (p^i) such that the hamiltonian dynamics starting from any (q_0,p_i) reach approximately (q_1,p^i_1) at time t=1, the shooting problem is performed by minimizing the LDDMM loss defined previously

    Parameters
    ----------

        q0 : array of shape (n_samples,d+1)
            initial time series' graph
        q0_mask : array of shape (n_samples,1)
            initial time series' graph mask
        batched_q1 : array of shape (N_timeseries,batch_size,n_samples,d+1)
            target time series' graphs, in practice batch_size is set to 1
        batched_q1_mask : array of shape (N_timeseries,batch_size,n_samples,1)
            target time series' graphs mask
        Kv : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        dataloss: (X,X_mask,Y,Y_mask)->float
            the attachment loss function, typically a varifold loss
        p0 : array of shape (n_samples,d+1)
            initial time series' graph momentum (facultative)
        gamma_loss: float 
            regularization constant related to the norm of the initial velocity
        niter : int
            number of iterations for the optimizer
        optimizer : optimize object from optax
            Default to optax.adabelief(learning_rate=0.1)
        nt : int
            number of integration time points, default=10
        delta : float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt, default to 1
        verbose : Boolean
            Default to True, print loss every 10 iterations
        
    Returns 
    -------

        batched_p : array of shape (n_samples,d+1)
            momentums minimizing the LDDMM loss related to the multiple shooting problems
        q0 : array of shape (n_samples,d+1)
            initial time series' graph
        q0_mask : array of shape (n_samples,1)
            initial time series' graph mask
    """
    loss = LDDMMLoss(Kv,dataloss,gamma_loss,nt,deltat)
    opt = BatchOneToManyRegistrationOptimizer(loss,niter,optimizer,verbose=verbose,stream_bool=stream_bool,stream_object=stream_object)
    if batched_p0 is None:
        batched_p0 = jnp.zeros_like(batched_q1)
    if callable(batched_p0):
        batched_p0 = batched_p0(q0,q0_mask,batched_q1,batched_q1_mask)
    batched_p = opt(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask)
    return batched_p,q0,q0_mask



# same should be used within a class
def batch_one_to_many_varifold_registration(q0,q0_mask,batched_q1,batched_q1_mask,Kv,Kl,batched_p0=None,gamma_loss=0.0,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=True,stream_bool=True,stream_object=None):
    """
    Return the momentums (p^i) such that the hamiltonian dynamics starting from any (q_0,p_i) reach approximately (q_1,p^i_1) at time t=1, the shooting problem is performed by minimizing the LDDMM loss defined previously

    Parameters
    ----------

        q0 : array of shape (n_samples,d+1)
            initial time series' graph
        q0_mask : array of shape (n_samples,1)
            initial time series' graph mask
        batched_q1 : array of shape (N_timeseries,batch_size,n_samples,d+1)
            target time series' graphs, in practice batch_size is set to 1
        batched_q1_mask : array of shape (N_timeseries,batch_size,n_samples,1)
            target time series' graphs mask
        Kv : (X,mask_X,Y,mask_Y,b)-> shape of b
            The Velocity kernel function (ideally the one given by TS-LDDMM)
        Kl : (X,mask_X,Y,mask_Y,b)-> array of the shape of b
            The Varifold kernel function. :math:`K(X,Y)b` where X and Y are array of size (n_samples,d+1), :math:`K(X,Y)` is the kernel matrix :math:`(k(x_i,y_j))`
            and b is an array of shape (n_samples,d) with d the dimension of the problem
        p0 : array of shape (n_samples,d+1)
            initial time series' graph momentum (facultative)
        gamma_loss: float 
            regularization constant related to the norm of the initial velocity
        niter : int
            number of iterations for the optimizer
        optimizer : optimize object from optax
            Default to optax.adabelief(learning_rate=0.1)
        nt : int
            number of integration time points, default=10
        delta : float 
            total time of integration :math:`\\delta`, stepsize=deltat/nt, default to 1
        verbose : Boolean
            Default to True, print loss every 10 iterations
        
    Returns 
    -------

        batched_p : array of shape (n_samples,d+1)
             initial momentums minimizing the LDDMM loss related to the multiple shooting problems
        q0 : array of shape (n_samples,d+1)
            initial time series' graph
        q0_mask : array of shape (n_samples,1)
            initial time series' graph mask
    """
    dataloss = VarifoldLoss(Kl)
    return batch_one_to_many_registration(q0,q0_mask,batched_q1,batched_q1_mask,Kv,dataloss,batched_p0,gamma_loss,niter,optimizer,nt,deltat,verbose,stream_bool=stream_bool,stream_object=stream_object)
# def Flowing(K:callable,nt=10,deltat=1.): 
#     Fri = FlowRalstonIntegrator(K,nt,deltat)
#     def flow(x0,p0,q0,q0_mask): 
#         return Fri(x0,p0,q0,q0_mask)
#     return flow
#def Shooting(K:callable,nt=10,deltat=1.0): 
#     """
#     - K : kernel function (timeseries graph,mask)^2x(moment,mask)->(velocity,mask)
#     - nt: number of time points 
#     - delta: total time of integration, stepsize= delta/nt

#     Return a function that tale into account initial position and momentum for Exp(1) shooting
#     """ 
#     Hri = HamiltonianRalstonIntegrator(K,nt,deltat)
#     def shoot(p0,q0,q0_mask): 
#         return Hri(p0,q0,q0_mask)
#     return shoot

####################################################################################################################################
####################################################################################################################################
### TIME ###
####################################################################################################################################
####################################################################################################################################

# def TimeLDDMMLoss(K:callable,dataloss:callable,gamma =0.001,nt=10,deltat=1.0): 
#     Hm = Hamiltonian(K)
#     shoot =Shooting(K,nt,deltat)
#     def loss(p0,q0,q0_mask,q1,q1_mask): 
#         t_q0, s_q0 = q0[:,:1],q0[:,1:]
#         p,t_q = shoot(p0,t_q0,q0_mask)
#         return gamma * Hm(p0,t_q0,q0_mask) + dataloss(jnp.hstack((t_q,s_q0)),q0_mask,q1,q1_mask)
#     return loss


# def TimeShooting(K:callable,nt=10,deltat=1.0): 
#     Hri = HamiltonianRalstonIntegrator(K,nt,deltat)
#     def shoot(p0,q0,q0_mask): 
#         t_q0,s_q0 = q0[:,:1],q0[:,1:]
#         p,t_q =Hri(p0,t_q0,q0_mask)
#         q = jnp.hstack((t_q,s_q0))
#         return p,q
#     return shoot

# def TimeFlowing(K:callable,nt=10,deltat=1.): 
#     Fri = FlowRalstonIntegrator(K,nt,deltat)
#     def flow(x0,p0,q0,q0_mask): 
#         t_x0,s_x0 = x0[:,:1],x0[:,1:]
#         t_q0,s_q0 = q0[:,:1],q0[:,1:]
#         t_x,p,t_q = Fri(t_x0,p0,t_q0,q0_mask)
#         x = jnp.hstack((t_x,s_x0))
#         q = jnp.hstack((t_q,s_q0))
#         return x,p,q
#     return flow

# def TimeDeformationGradient(K:callable,nt=10,deltat=1.0): 
#     jacHri = jacfwd(HamiltonianRalstonIntegrator(K,nt,deltat),1)
#     def gradient(p0,q0,q0_mask): 
#         t_q0,s_q0 = q0[:,:1],q0[:,1:]
#         gr =jacHri(p0,t_q0,q0_mask)[1]
#         return np.diag(gr[:,0,:,0]).reshape(-1,1)-1
#     return gradient

# def time_registration(q0,q0_mask,q1,q1_mask,Kv,dataloss,gamma_loss=0.001,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=True):
#     loss = TimeLDDMMLoss(Kv,dataloss,gamma_loss,nt,deltat)
#     opt = RegistrationOptimizer(loss,niter,optimizer,verbose)
#     p0 = jnp.zeros_like(q0[:,:1])
#     p = opt(p0,q0,q0_mask,q1,q1_mask)
#     return p,q0,q0_mask

# def varifold_time_registration(q0,q0_mask,q1,q1_mask,Kv,Kl,gamma_loss=0.001,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=True):
#     dataloss = VarifoldLoss(Kl)
#     return time_registration(q0,q0_mask,q1,q1_mask,Kv,dataloss,gamma_loss,niter,optimizer,nt,deltat,verbose)

# def batch_one_to_many_time_registration(q0,q0_mask,batched_q1,batched_q1_mask,Kv,dataloss,batched_p0=None,gamma_loss=0.0,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=True):
#     loss = TimeLDDMMLoss(Kv,dataloss,gamma_loss,nt,deltat)
#     opt = BatchOneToManyRegistrationOptimizer(loss,niter,optimizer,verbose=verbose)
#     if batched_p0 is None:
#         batched_p0 = jnp.zeros_like(batched_q1[:,:,:,:1])
#     batched_p = opt(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask)
#     return batched_p,q0,q0_mask

# def batch_one_to_many_varifold_time_registration(q0,q0_mask,batched_q1,batched_q1_mask,Kv,Kl,batched_p0=None,gamma_loss=0.0,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=True):
#     dataloss = VarifoldLoss(Kl)
#     return batch_one_to_many_time_registration(q0,q0_mask,batched_q1,batched_q1_mask,Kv,dataloss,batched_p0,gamma_loss,niter,optimizer,nt,deltat,verbose)

# def time_initializer(Kv,dataloss,gamma_loss=0.0,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=False):
#     def initializer(q0,q0_mask,q1,q1_mask):
#         p,q,qm = time_registration(q0,q0_mask,q1,q1_mask,Kv,dataloss,None,gamma_loss,niter,optimizer,nt,deltat,verbose)
#         return jnp.pad(p,((0,0),(0,q.shape[1]-1)))
#     return initializer

# def varifold_time_initializer(Kv,Kl,gamma_loss=0.0,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=False):
#     def initializer(q0,q0_mask,q1,q1_mask):
#         p,q,qm = varifold_time_registration(q0,q0_mask,q1,q1_mask,Kv,Kl,gamma_loss,niter,optimizer,nt,deltat,verbose)
#         return jnp.pad(p,((0,0),(0,q.shape[1]-1)))
#     return initializer

# def batch_time_initializer(Kv,dataloss,gamma_loss=0.0,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=False):
#     def initializer(q0,q0_mask,batched_q1,batched_q1_mask):
#         bp,bq,bqm = batch_one_to_many_time_registration(q0,q0_mask,batched_q1,batched_q1_mask,Kv,dataloss,None,gamma_loss,niter,optimizer,nt,deltat,verbose)
#         return jnp.pad(bp,((0,0),(0,0),(0,0),(0,bq.shape[1]-1)))
#     return initializer

# def batch_varifold_time_initializer(Kv,Kl,gamma_loss=0.0,niter=100,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,verbose=False):
#     def initializer(q0,q0_mask,batched_q1,batched_q1_mask):
#         bp,bq,bqm = batch_one_to_many_varifold_time_registration(q0,q0_mask,batched_q1,batched_q1_mask,Kv,Kl,None,gamma_loss,niter,optimizer,nt,deltat,verbose)
#         return jnp.pad(bp,((0,0),(0,0),(0,0),(0,bq.shape[1]-1)))
#     return initializer









import numpy as np
from .utils import time_shape_embedding




def generate_easy_dataset(N=8):
    Dataset=np.zeros(((N-1)**2,200,2))
    Dataset_mask=np.zeros(((N-1)**2,200,1))
    Feature=np.zeros(((N-1)**2,2))
    for i in range(1,N):
        for j in range(1,N):
            t_sig_new=np.hstack((np.sin(np.linspace(0,0.5,j*10-1)*2*np.pi),np.sin(np.linspace(0.5,1,200-j*10+1)*2*np.pi)))
            t_sig_new[150:170] += i*2*np.sin(np.linspace(0,1,20)*2*np.pi)/20
            t_sig_new = t_sig_new.reshape(-1,1)
            t_sig_new=time_shape_embedding(t_sig_new)
            Dataset[(i-1)*(N-1)+j-1,:,:] = t_sig_new.copy()
            Dataset_mask[(i-1)*(N-1)+j-1,:,:] = np.full_like(t_sig_new[:,:1],True).astype(bool)
            Feature[(i-1)*(N-1)+j-1]=np.array([i,j])
    #n_ts,n_s,n_d = dataset.shape

    graph_ref=Dataset[(N//2-1)*(N-1)+N//2-1,:,:]
    graph_ref_mask=Dataset_mask[(N//2-1)*(N-1)+N//2-1,:,:]
    return Dataset,Dataset_mask,graph_ref,graph_ref_mask




import optax
from jax import jit,grad,value_and_grad,vmap
import jax.numpy as jnp
from jax import lax
import numpy as np
from functools import partial


####################################################################################################################################
####################################################################################################################################
### Registration ###
####################################################################################################################################
####################################################################################################################################


def RegistrationOptimizer(loss:callable,niter=100,optimizer = optax.adam(learning_rate=0.1),verbose=True,stream_bool=False,stream_object=None):
    
    if stream_bool:
        bar,latest_iteration=stream_object
    def f(p0,q0,q0_mask,q1,q1_mask):

        p=p0
        opt_state = optimizer.init(p)
        jit_loss = jit(value_and_grad(loss))

        def step(p,q0,q0_mask,q1,q1_mask,opt_state):
            loss_value,grad = jit_loss(p,q0,q0_mask,q1,q1_mask) 
            updates, opt_state = optimizer.update(grad, opt_state,p)
            p = optax.apply_updates(p,updates)
            return p, opt_state, loss_value
        
        for i in range(niter): 
            p,opt_state,loss_value = step(p,q0,q0_mask,q1,q1_mask,opt_state)
            if stream_bool:
                bar.progress((i + 1)*100//niter)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    if stream_bool:
                        latest_iteration.text(f"Iteration: {i+1} niter -- loss:"+"{:0.2f}".format(loss_value))
    
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))
        return p
    return f 

def BatchOneToManyRegistrationOptimizer(loss:callable,niter=100,optimizer = optax.adam(learning_rate=0.1),verbose=True,stream_bool=False,stream_object=None):
    if stream_bool:
        bar,latest_iteration=stream_object
    def f(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask):

        n_batches,batch_size = batched_p0.shape[0],batched_p0.shape[1]
        batched_p=batched_p0
        opt_state = optimizer.init(batched_p)
        vmap_loss = vmap(loss, (0,None,None,0,0),0)
        def rloss(p0,q0,q0_mask,q1,q1_mask): 
            return jnp.sum(vmap_loss(p0,q0,q0_mask,q1,q1_mask))
        jit_loss = jit(value_and_grad(rloss))

        def step(batched_p,q,q_mask,batched_q1,batched_q1_mask,opt_state):

            p_grad = np.zeros_like(batched_p)
            loss_value = 0

            for i,(p,q1,q1_mask) in enumerate(zip(batched_p,batched_q1,batched_q1_mask)):
                t_loss_value,grads = jit_loss(p,q,q_mask,q1,q1_mask)
                p_grad[i] = grads
                loss_value += t_loss_value

           
            loss_value /= (n_batches*batch_size)
            updates, opt_state = optimizer.update(p_grad, opt_state,batched_p)
            batched_p = optax.apply_updates(batched_p,updates)

            return batched_p, opt_state, loss_value
        
        for i in range(niter): 
            if stream_bool:
                bar.progress((i + 1)*100//niter)
            batched_p,opt_state,loss_value = step(batched_p,q0,q0_mask,batched_q1,batched_q1_mask,opt_state)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))
                    if stream_bool:
                        latest_iteration.text(f"Iteration: {i+1} niter -- loss:"+"{:0.2f}".format(loss_value))
        return batched_p
    return f 


####################################################################################################################################
####################################################################################################################################
### Barycenter ###
####################################################################################################################################
####################################################################################################################################

# Explain this ootmizer
def BatchIteratedBarycenterOptimizer(shoot:callable,loss:callable,niter=200, optimizer = optax.adam(learning_rate=0.1), update_interval =100,verbose=True,stream_bool=None,stream_object=None):
    if stream_bool:
        bar,latest_iteration=stream_object
    def f(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask):
        
        n_batches,batch_size = batched_p0.shape[0],batched_p0.shape[1]
        batched_p,q = batched_p0, q0
        opt_state = optimizer.init((batched_p))
        jit_loss = jit(value_and_grad(loss))

        def step(batched_p,q,q_mask,batched_q1,batched_q1_mask,opt_state):

            p_grad = np.zeros_like(batched_p)
            loss_value = 0

            for i,(p,q1,q1_mask) in enumerate(zip(batched_p,batched_q1,batched_q1_mask)):
                t_loss_value,grads = jit_loss(p,q,q_mask,q1,q1_mask)
                p_grad[i] = grads[0]
                loss_value += t_loss_value

            loss_value /= (n_batches*batch_size)
            updates, opt_state = optimizer.update(p_grad, opt_state,batched_p)
            batched_p = optax.apply_updates(batched_p,updates)

            return batched_p, opt_state, loss_value
        

        for i in range(niter): 
            if stream_bool:
                bar.progress((i + 1)*100//niter)
            batched_p,opt_state,loss_value = step(batched_p,q,q0_mask,batched_q1,batched_q1_mask,opt_state)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))
                    if stream_bool:
                        latest_iteration.text(f"Iteration: {i+1} niter -- loss:"+"{:0.2f}".format(loss_value))
                if ((i+1)!=niter)*(i!=0)*((i+1)%update_interval == 0):
                    print("...Updating barycenter...")
                    pbar = np.mean(batched_p,axis=(0,1))
                    _,q = shoot(pbar,q,q0_mask)
                    batched_p = batched_p - pbar[None,:,:]        
        return batched_p,q
    return f 







import numpy as np
import jax.numpy as jnp
from jax import vmap,jit
from jax.lax import map 
from sklearn.base import BaseEstimator
from numpy.linalg import svd
from scipy.linalg import solve
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from functools import partial
from scipy.linalg import sqrtm,inv


# Will not change that much, we can doc it



def MaskMomentaGramMatrix(K,qbar,qbar_mask): # return a function that generate the gram matrix of an array of momentums
    f = lambda u,v : jnp.sum(u*K(qbar,qbar_mask,qbar,qbar_mask,v))
    f = vmap(f,(0,None),0)
    f = vmap(f,(None,0),1)
    return jit(f)






class MomentaPCA(BaseEstimator):

    def __init__(self,n_component,centered=False,type = "cov") -> None:
        super().__init__()
        self.n_component = n_component
        self.centered = centered
        self.type = type    


    def fit(self,Kv,ps,qbar,qbar_mask): 

        self.Kv_ = Kv
        self.qbar_ = qbar
        self.ps_ = ps

    
        self._mgm = MaskMomentaGramMatrix(self.Kv_,qbar,qbar_mask)
       

        if self.centered:
            r_ps = ps
        else:
            self.m_ps_ = np.mean(ps,axis=0)
            r_ps = ps - self.m_ps_[None,:,:]

        X = self._mgm(r_ps,r_ps)

        if self.type == "cor": 
            inv_norm = 1/np.sqrt(np.diag(X))
            X= X*(inv_norm.reshape(-1,1)*inv_norm.reshape(1,-1))
        
        U,S,Vh = svd(X,full_matrices=True,compute_uv=True,hermitian=True)

        eig_vec = U[:,:self.n_component]/np.sqrt(S[:self.n_component]).reshape(1,-1)
        self.p_pc_= np.sum(eig_vec.T[:,:,None,None] * r_ps[None,:,:,:],axis=1)
        self.p_std_ = np.sqrt(S[:self.n_component]/(X.shape[0]-1))
        self.p_score_ = X @ eig_vec
        
    def transform(self,ps): 
        if self.centered:
            r_ps = ps
        else:
            r_ps = ps - self.m_ps_[None,:,:]
        return self._mgm(r_ps,self.p_pc_)
    

class MomentaLDA(LinearDiscriminantAnalysis):

    def __init__(self, solver = "svd", shrinkage = None, priors = None, n_components = None, store_covariance = False, tol = 0.0001, covariance_estimator = None) -> None:
        super().__init__(solver, shrinkage, priors, n_components, store_covariance, tol, covariance_estimator)

    def _fit_initialization(self,Kv,ps,qbar,qbar_mask): 
        self.Kv_ = Kv
        self.qbar_ = qbar
        self.ps_ = ps
        self._kgm = MaskMomentaGramMatrix(self.Kv_,qbar,qbar_mask)
        self.G_ = self._kgm(self.ps_,self.ps_)
        #U,S,Vh = svd(self.G_,full_matrices=True,compute_uv=True,hermitian=True)
        #self.s_G_ = U @ (np.sqrt(S).reshape(-1,1) * Vh)
        #self.inv_s_G_ = U @ (1/np.sqrt(S).reshape(-1,1)* Vh)
        self.s_G_ = sqrtm(self.G_)
        self.inv_s_G_ = sqrtm(inv(self.G_))

    def fit(self,Kv,ps,qbar,qbar_mask,y):
        self._fit_initialization(Kv,ps,qbar,qbar_mask)
        super().fit(self.s_G_,y)
        self.p_coef_ = (self.coef_/np.linalg.norm(self.coef_,axis=1).reshape(-1,1)) @ self.inv_s_G_
        self.p_lda_ = np.sum(self.p_coef_[:,:,None,None] * self.ps_[None,:,:,:],axis=1)
        self.p_score_ = self.p_coef_ @ self.G_

    def _embedding(self, ps): 
        if ps.ndim ==2: 
            ps = ps[None,:,:]
        embeded =self._kgm(ps,self.ps_) @ self.inv_s_G_ 
        return embeded

    def predict(self,ps):
        embeded = self._embedding(ps)
        return super().predict(embeded)
    
    def predict_proba(self,ps): 
        embeded = self._embedding(ps)
        return super().predict_proba(embeded)

    def predict_log_proba(self, ps):
        embeded = self._embedding(ps)
        return super().predict_log_proba(embeded)
    
    def predict_p_score(self,ps): 
        return self._kgm(ps,self.p_lda_)



import numpy as np
from .kernel import Varifold_TSLDDMM_Gaussian_Kernel,Velocity_TSLDDMM_Gaussian_Kernel
from .lddmm import Shooting, batch_one_to_many_registration
from .utils import time_shape_embedding,batch_dataset
from .loss import SumVarifoldLoss
from .statistic import MomentaPCA
from .barycenter import batch_barycenter_registration
from optax.schedules import warmup_cosine_decay_schedule
import optax
import matplotlib.pyplot as plt

#TODO preprocessing, from_timeseries_to_dataset, tuto, synthetic




#kernel_velocity
#c_{time}->c_0
#c_{space}->c_1
#\sigma_{time,0}->t_sigma_1
#\sigma_{time,1}->t_sigma_2
#\sigma_{space,1}->s_sigma

#kernel varifold
# "$\sigma_{pos,t}$", min_value =0.1, max_value =4.0, step=0.1,value=2.0)->t_sigma_1
# "$\sigma_{pos,x}$", min_value =0.1, max_value =4.0, step=0.1,value=1.0)->s_sigma_1
# $\sigma_{dir,t}$", min_value =0.1, max_value =4.0, step=0.1,value=2.0)->t_sigma_2
# $\sigma_{dir,x}$", min_value =0.1, max_value =4.0,step= 0.1,value=0.6)->s_sigma_2

Defaut_velocity_hyper={"c_0":1.0,"c_1":0.1,"t_sigma_1":60,"t_sigma_2":1.0,"s_sigma":1.0}
Defaut_varifold_hyper={"t_sigma_1":2.0,"s_sigma_1":1.0,"t_sigma_2":2.0,"s_sigma_2":0.6}

niter_defaut=100
schedule_defaut =warmup_cosine_decay_schedule(0,0.3,int(0.1*niter_defaut),niter_defaut,0)
#schedule =warmup_cosine_decay_schedule(0,0.3,80,800,0)


class TS_PCA_():

    def __init__(self,kernel_velocity_arg=Defaut_velocity_hyper,kernel_varifold_arg=Defaut_varifold_hyper):
        """
        Initialize the TS-PCA object with velocity and varifold kernels for TS-LDDMM.

        The velocity kernel is a Gaussian kernel defined as in "Shape analysis for time series":
        :math:`K_G((t,x),(t',x')) = \begin{pmatrix} c_0 K_{\text{time}}(t,t') & 0 \\ 0 & c_1 K_{\text{space}}((t,x),(t',x')) \end{pmatrix}`,
        where :math:`K_{\text{time}}` and :math:`K_{\text{space}}` are isotropic Gaussian kernels with parameters specified in `kernel_velocity_arg`.

        The varifold kernel is a Gaussian kernel used to compute distances between shapes along the time series,
        with parameters specified in `kernel_varifold_arg`.

        Parameters
        ----------
            kernel_velocity_arg : dict, optional
                Hyperparameters for the velocity kernel (default: `Defaut_velocity_hyper`):
                - `c_0` : float
                    Time scaling coefficient.
                - `c_1` : float
                    Space scaling coefficient.
                - `t_sigma_1` : float
                    Variance parameter for the temporal component (sigma T0).
                - `t_sigma_2` : float
                    Variance parameter for the temporal component (sigma T1).
                - `s_sigma` : float
                    Variance parameter for the spatial component (sigma_x).

            kernel_varifold_arg : dict, optional
                Hyperparameters for the varifold kernel (default: `Defaut_varifold_hyper`):
                - `t_sigma_1`, `t_sigma_2` : float
                    Variance parameters for the temporal components.
                - `s_sigma_1`, `s_sigma_2` : float
                    Variance parameters for the spatial components.

        Attributes
        ----------
            velocity_kernel : Velocity_TSLDDMM_Gaussian_Kernel
                Gaussian kernel used for the velocity in TS-LDDMM.
            varifold_kernel : Varifold_TSLDDMM_Gaussian_Kernel
                Gaussian kernel used for computing varifold distances.
            shoot : Shooting
                Shooting object initialized with the velocity kernel for TS-LDDMM computations.
            fit_TS_LDDMM : bool
                Flag indicating whether TS-LDDMM representations have been fitted (initialized as False).

        """
        
        
        self.velocity_kernel=Velocity_TSLDDMM_Gaussian_Kernel(kernel_velocity_arg["c_0"],kernel_velocity_arg["c_1"],kernel_velocity_arg["t_sigma_1"],kernel_velocity_arg["t_sigma_2"],kernel_velocity_arg["s_sigma"])
        self.varifold_kernel=Varifold_TSLDDMM_Gaussian_Kernel(kernel_varifold_arg["t_sigma_1"],kernel_varifold_arg["s_sigma_1"],kernel_varifold_arg["t_sigma_2"],kernel_varifold_arg["s_sigma_2"])
        self.shoot= Shooting(self.velocity_kernel)
        self.fit_TS_LDDMM=False


    def fit_TS_LDDMM_representations(self,Dataset,Dataset_mask,niter=niter_defaut,schedule=schedule_defaut,gamma_loss=1e-3,learning_graph_ref=True,graph_ref=None,graph_ref_mask=None,dataloss=None):
        """
        Fit TS-LDDMM representations of a dataset of time series graphs.

        This method computes the TS-LDDMM embeddings of the input dataset by optimizing 
        initial momentums (:math:`p_0`) and optionally the reference graph 
        (:math:`G_0`) using a velocity kernel and a varifold data term. 
        If `learning_graph_ref` is True, the reference graph is estimated via a barycenter 
        procedure; otherwise, a fixed reference graph must be provided.

        Parameters
        ----------
            Dataset : array of shape (n_time_series, n_samples_max, d+1)
                Input time series graphs. The first column corresponds to time points, 
                the remaining columns are spatial values of dimension d.

            Dataset_mask : array of shape (n_time_series, n_samples_max, 1)
                Mask indicating valid samples in each time series graph.

            niter : int, optional
                Number of iterations for the optimization (default: `niter_defaut`).

            schedule : object, optional
                Learning rate schedule for the optax optimizer (default: `schedule_defaut`).

            gamma_loss : float, optional
                Weighting factor for the data loss term in the optimization (default: 1e-3).

            learning_graph_ref : bool, optional
                If True, the reference graph `graph_ref` is estimated jointly with momentums;
                if False, a fixed reference graph must be provided (default: True).

            graph_ref : array of shape (n_samples_max, d+1), optional
                Reference graph to use if `learning_graph_ref=False`. Must be provided together 
                with `graph_ref_mask`.

            graph_ref_mask : array of shape (n_samples_max, 1), optional
                Mask for the reference graph.

            dataloss : object, optional
                Data loss object used for registration. If None, a default `SumVarifoldLoss` 
                with the internal varifold kernel is used.

        Returns
        -------
            momentums : array
                Optimized initial momentums (:math:`p_0`) for the TS-LDDMM mapping.

            graph_ref : array
                Learned or provided reference graph (:math:`G_0`) after optimization.

            graph_ref_mask : array
                Mask for the reference graph.

        Attributes
        ----------
            fit_TS_LDDMM_bool : bool
                Flag indicating that TS-LDDMM representations have been fitted successfully 
                (set to True at the end of the method).

        Notes
        -----
        - This function internally uses batching via `batch_dataset` and the `Shooting` object.
        - Optimization is performed using the Adabelief optimizer from `optax`.
        - Raises a ValueError if `learning_graph_ref=False` and `graph_ref` or `graph_ref_mask` are not provided.
        """
        bDataset,bDataset_mask = batch_dataset(Dataset,1,Dataset_mask)
        optimize=optax.adabelief(schedule)

        if dataloss is None:
            dataloss=SumVarifoldLoss([self.varifold_kernel])
        if not(learning_graph_ref) :
            if graph_ref is None or graph_ref_mask is None:
                raise ValueError("You should give a graph of reference 'graph_ref' array(n_samples,d+1) and its mask 'graph_ref_mask' array(n_samples,1)")
            self.graph_ref=graph_ref# G_0
            self.graph_ref_mask=graph_ref_mask
            print("run momentums learning")
            p0s,q0,q0_mask= batch_one_to_many_registration(self.graph_ref,self.graph_ref_mask,bDataset,bDataset_mask,self.velocity_kernel,dataloss,niter=niter,optimizer=optimize,gamma_loss=gamma_loss,stream_bool=False)
            self.momentums=p0s
        else:
            print("run momentums+graph_ref learning")
            p0s,q0,q0_mask = batch_barycenter_registration(self.shoot,bDataset,bDataset_mask,self.velocity_kernel,dataloss,niter=niter,optimizer=optimize,gamma_loss=gamma_loss,stream_bool=False)
            self.momentums=p0s
            self.graph_ref=q0# G_0
            self.graph_ref_mask=q0_mask
        self.fit_TS_LDDMM_bool=True
        return self.momentums,self.graph_ref,self.graph_ref_mask



    def fit_kernel_PCA(self,n_comp=3,momentums=None,graph_ref=None,graph_ref_mask=None):
        """
        Fit a Kernel PCA on TS-LDDMM momentums to extract principal components.

        This method performs Kernel Principal Component Analysis (KPCA) on the initial 
        momentums (:math:`p_0`) obtained from TS-LDDMM representations. It computes 
        a low-dimensional embedding of the momentums, capturing the main modes of 
        variation in the dataset.

        If `momentums` are not provided, the method uses the momentums, reference graph,
        and reference graph mask stored in the object (from a previous call to 
        `fit_TS_LDDMM_representations`).

        Parameters
        ----------
            n_comp : int, optional
                Number of principal components to extract (default: 3).

            momentums : array of shape (n_ind, n_samples, n_dim+1), optional
                TS-LDDMM initial momentums (:math:`p_0`). If None, the method uses the 
                internal momentums stored in the object.

            graph_ref : array of shape (n_samples, d+1), optional
                Reference graph (:math:`G_0`) corresponding to the momentums. Required if 
                `momentums` is provided.

            graph_ref_mask : array of shape (n_samples, 1), optional
                Mask for the reference graph. Required if `momentums` is provided.

        Returns
        -------
            None

        Attributes
        ----------
            n_comp : int
                Number of components used for KPCA.

            mean_ : array
                Mean momentum (:math:`\\bar{p}_0`) computed across samples.

            singular_values_ : array
                Standard deviation of principal components.

            components_ : array
                Principal components matrix, each row corresponds to a component.

            n_samples_ : int
                Number of time series in the dataset.

            coordinates_ : array
                Projected coordinates of each momentum in the KPCA space.

            fit_kernel_PCA_bool : bool
                Flag indicating that Kernel PCA has been successfully fitted (set to True).

        Raises
        ------
            ValueError
                If `momentums`, `graph_ref`, or `graph_ref_mask` are not provided and 
                cannot be inferred from the object. Suggests using 
                `fit_TS_LDDMM_representations` to generate these inputs.

        Notes
        -----

            - The method internally uses `MomentaPCA` with covariance-based KPCA.
            - Each momentum is reshaped to match the expected input dimensions for KPCA.
            - Kernel PCA captures the main modes of variation of TS-LDDMM momentums,
            which can then be used for visualization or downstream analysis.

        """
        if self.fit_TS_LDDMM_bool and momentums is None:
            momentums,graph_ref,graph_ref_mask=self.momentums,self.graph_ref,self.graph_ref_mask
            momentums = momentums.reshape(-1,momentums.shape[2],momentums.shape[3])
        
        if momentums is None or graph_ref is None or graph_ref_mask is None:
            raise ValueError("You should give momentums (n_ind,n_samples,n_dim+1) , a graph of reference 'graph_ref' array(n_samples,d+1) and its mask 'graph_ref_mask' array(n_samples,1). You should use 'fit_TS_LDDMM_representations' to get it. ")
        else:
            print("run kernel PCA on the momentums")
            mpca = MomentaPCA(n_comp,False,"cov")
            
            mpca.fit(self.velocity_kernel,momentums,graph_ref,graph_ref_mask)
        self.n_comp=n_comp
        self.mean_ = mpca.m_ps_ #mean p0
        self.singular_values_=mpca.p_std_
        self.components_=mpca.p_pc_ # each line
        self.n_samples_=len(momentums)
        self.coordinates_=mpca.p_score_
        self.fit_kernel_PCA_bool=True


    def fit_transform(self,n_comp=2,momentums=None,graph_ref=None,graph_ref_mask=None):
        """

        Apply fit_kernel_PCA and return the Projected coordinates of each momentum in the Kernel PCA space.

        """
        self.fit_kernel_PCA(self,n_comp=n_comp,momentums=momentums,graph_ref=graph_ref,graph_ref_mask=graph_ref_mask)
        return self.coordinates_
    
    def plot_components(self):
        """

        Plot the the effect of the components of Kernel PCA as deformations on the reference graph

        """

        if not(self.fit_kernel_PCA_bool):
            raise ValueError(" You should first call 'fit_kernel_PCA' to get the components. ")

        else:
            n_disp=3
            ft_size = 25
            pft_size = 20
            components = [f"PC{i+1}" for i in range(self.n_comp)]
            p0_b = self.mean_
            plot_mask = np.where(self.graph_ref_mask==True)[0]
            fig2,axs = plt.subplots(self.n_comp,n_disp,figsize = (n_disp*3,self.n_comp*3),sharex=True,sharey=True)
            for pca_index in range(self.n_comp):
                sigma_pca = self.singular_values_[pca_index]
                for j,alpha in enumerate(np.linspace(-2*sigma_pca , 2*sigma_pca, n_disp)):
                    p0_mode = p0_b + alpha * self.components_[pca_index]
                    _,q = self.shoot(p0_mode,self.graph_ref,self.graph_ref_mask)
                    plot_q = q[plot_mask].T
                    axs[pca_index,j].plot(plot_q[0],plot_q[1],color="tab:red")
                    #axs[pca_index,j].plot(plot_q[0],plot_q[2])
                axs[pca_index,0].set_ylabel(components[pca_index],fontsize=ft_size)



            for ax in axs.flatten():
                ax.axhline(0,color="black",linewidth=1,linestyle="--",zorder=0)

            axs[-1,1].set_xlabel("time (ms)",fontsize=pft_size)

            axs[0,0].set_title(r"-2$\sigma_{PC}$",fontsize=ft_size)
            axs[0,1].set_title(r"$\mathsf{G}_0$",fontsize=ft_size)
            axs[0,2].set_title(r"+2$\sigma_{PC}$",fontsize=ft_size)
            axs[0,2].legend(fontsize = pft_size,loc=1)
            for ax in axs[-1,:]:
                ax.xaxis.set_tick_params(labelsize=pft_size)
            for ax in axs[:,0]:
                ax.yaxis.set_tick_params(labelsize=pft_size)

            #fig.suptitle("II) PC shootings",fontsize = ft_size)
            fig2.suptitle(" ",fontsize = ft_size)

            fig2.subplots_adjust(left=0.1,bottom=0.15,right=0.9,top=0.85,wspace=0.1,hspace=0.15)
            fig2.tight_layout()

            plt.title("kernelPCA components related to the TS-LDDMM representation")
            plt.show()


import numpy as np
import optax

from sklearn.svm import SVC
from sklearn.metrics import accuracy_score,f1_score
from .statistic import MaskMomentaGramMatrix
from .utils import batch_dataset,unbatch_dataset
from .lddmm import batch_one_to_many_registration
from .barycenter import batch_barycenter_registration


class MomentaSVC(SVC):

    def __init__(self,Kv,dataloss,barycenter_initializer = None,registration_initializer=None,C=1,max_per_batch =10,gamma_loss=0.0,niter=200,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0, shrinking = True, probability = False, tol = 0.001, cache_size = 200, class_weight = None, verbose = True, max_iter =-1, decision_function_shape = "ovr", break_ties = False, random_state = None) -> None:
        super().__init__(C=C, kernel="precomputed",shrinking=shrinking, probability=probability, tol=tol, cache_size=cache_size, class_weight=class_weight, verbose=False, max_iter=max_iter, decision_function_shape=decision_function_shape, break_ties=break_ties, random_state=random_state)
        
        self.Kv = Kv
        self.dataloss = dataloss
        self.barycenter_initializer = barycenter_initializer
        self.registration_initializer = registration_initializer
        self.max_per_batch = max_per_batch
        self.gamma_loss = gamma_loss
        self.niter = niter
        self.optimizer = optimizer
        self.nt = nt
        self.deltat = deltat
        self.verbose = verbose

    def _batch(self,X,X_mask): 
        n_samples = X.shape[0]
        for i in np.arange(self.max_per_batch,0,-1): 
            if n_samples % i ==0: 
                self.batch_size_ = i
                batched_X, batched_X_mask = batch_dataset(X,self.batch_size_,X_mask)
                break
        return batched_X, batched_X_mask
    
    def _unbached(self,batched_X, batched_X_mask=None): 
        return unbatch_dataset(batched_X,batched_X_mask)
    
    def fit(self,X,X_mask,y,init=None,init_mask=None): 
        batched_X,batched_X_mask = self._batch(X,X_mask)
        batched_ps,self.qbar_,self.qbar_mask_ = batch_barycenter_registration(batched_X,batched_X_mask,self.Kv,self.dataloss,init,init_mask,self.barycenter_initializer,self.niter,self.optimizer,self.gamma_loss,self.nt,self.deltat,self.verbose)
        self.ps_ = self._unbached(batched_ps)
        self._mgm = MaskMomentaGramMatrix(self.Kv,self.qbar_,self.qbar_mask_)
        self.X_svc_ = self._mgm(self.ps_,self.ps_)
        self.y_ = y
        super().fit(self.X_svc_,self.y_)
        return self
    
    def set_C(self,X,X_mask,y,C_lst):
        batched_X,batched_X_mask = self._batch(X,X_mask)
        batched_ps,_,_ = batch_one_to_many_registration(self.qbar_,self.qbar_mask_,batched_X,batched_X_mask,self.Kv,self.dataloss,self.registration_initializer,self.gamma_loss,self.niter,self.optimizer,self.nt,self.deltat,self.verbose)
        ps = self._unbached(batched_ps)
        X_svc = self._mgm(ps,self.ps_)
        best_C = 0
        best_score = -np.inf
        for C in C_lst: 
            self.C = C
            super().fit(self.X_svc_,self.y_)
            y_pred = super().predict(X_svc)
            score = f1_score(y,y_pred,average="macro")
            print(f"Validation -- C: {C} -- f1score: {score}")
            if score > best_score: 
                best_C = C
                best_score = score.copy()
        self.C = best_C
        print(f"Best C :{self.C}")
        return self
    
    def predict(self,X,X_mask):
        batched_X,batched_X_mask = self._batch(X,X_mask)
        batched_ps,_,_ = batch_one_to_many_registration(self.qbar_,self.qbar_mask_,batched_X,batched_X_mask,self.Kv,self.dataloss,self.registration_initializer,self.gamma_loss,self.niter,self.optimizer,self.nt,self.deltat,self.verbose)
        ps = self._unbached(batched_ps)
        X_svc = self._mgm(ps,self.ps_)
        return super().predict(X_svc)
    
    def score(self,X,X_mask,y):
        y_pred = self.predict(X,X_mask)
        return accuracy_score(y,y_pred)
    



from sklearn.svm import SVC,OneClassSVM
from sklearn.metrics import accuracy_score
from .utils import batch_dataset,unbatch_dataset
from .lddmm import batch_one_to_many_registration
from .barycenter import batch_barycenter_registration

class OneMomentaSVC(OneClassSVM):
    
    def __init__(self,Kv,dataloss,max_per_batch =10,gamma_loss=0.0,niter=200,optimizer = optax.adam(learning_rate=0.1),nt=10,deltat=1.0,degree = 3, gamma = "scale", coef0 = 0, tol = 0.001, nu = 0.5, shrinking =True, cache_size = 200, verbose=False, max_iter=-1) -> None:
        super().__init__(kernel="precomputed", degree=degree, gamma=gamma, coef0=coef0, tol=tol, nu=nu, shrinking=shrinking, cache_size=cache_size, verbose=verbose, max_iter=max_iter)

        self.Kv = Kv
        self.dataloss = dataloss
        self.max_per_batch = max_per_batch
        self.gamma_loss = gamma_loss
        self.niter = niter
        self.optimizer = optimizer
        self.nt = nt
        self.deltat = deltat
        self.verbose = verbose

    def _batch(self,X,X_mask): 
        n_samples = X.shape[0]
        for i in np.arange(self.max_per_batch,0,-1): 
            if n_samples % i ==0: 
                self.batch_size_ = i
                batched_X, batched_X_mask = batch_dataset(X,self.batch_size_,X_mask)
                break
        return batched_X, batched_X_mask
    
    def _unbached(self,batched_X, batched_X_mask=None): 
        return unbatch_dataset(batched_X,batched_X_mask)
    
    def fit(self,X,X_mask,y,init=None,init_mask=None): 
        idxs = y==1
        X = X[idxs]
        X_mask = X_mask[idxs]
        batched_X,batched_X_mask = self._batch(X,X_mask)
        batched_ps,self.qbar_,self.qbar_mask_ = batch_barycenter_registration(batched_X,batched_X_mask,self.Kv,self.dataloss,init,init_mask,self.niter,self.optimizer,self.gamma_loss,self.nt,self.deltat,self.verbose)
        self.ps_ = self._unbached(batched_ps)
        self._mgm = MaskMomentaGramMatrix(self.Kv,self.qbar_,self.qbar_mask_)
        X_svc = self._mgm(self.ps_,self.ps_)
        super().fit(X_svc)
        return self
    
    def predict(self,X,X_mask):
        batched_X,batched_X_mask = self._batch(X,X_mask)
        batched_ps,_,_ = batch_one_to_many_registration(self.qbar_,self.qbar_mask_,batched_X,batched_X_mask,self.Kv,self.dataloss,self.gamma_loss,self.niter,self.optimizer,self.nt,self.deltat,self.verbose)
        ps = self._unbached(batched_ps)
        X_svc = self._mgm(ps,self.ps_)
        return np.clip(super().predict(X_svc),0,1)
    
    def score(self,X,X_mask,y):
        y_pred = self.predict(X,X_mask)
        return accuracy_score(y,y_pred)


import optax
from jax import jit,grad,value_and_grad,vmap
import jax.numpy as jnp
from jax import lax
import numpy as np
from functools import partial


####################################################################################################################################
####################################################################################################################################
### GENERAL ###
####################################################################################################################################
####################################################################################################################################

def Optimizer(loss:callable,niter=100,optimizer = optax.adam(learning_rate=0.1),static_p0 =False, static_q0=True,verbose=True):
    def f(p0,q0,q0_mask,q1,q1_mask):
        
        if (static_p0==False)*(static_q0==True):
            arg = p0
            def map_loss(arg):
                return loss(arg,q0,q0_mask,q1,q1_mask)
        elif (static_p0==False)*(static_q0==False): 
            arg = (p0,q0)
            def map_loss(arg):
                return loss(*arg,q0_mask,q1,q1_mask)
        elif (static_p0==True)*(static_q0==False): 
            arg = q0
            def map_loss(arg): 
                return loss(p0,arg,q0_mask,q1,q1_mask)
            
        opt_state = optimizer.init(arg)
        jit_map_loss = jit(value_and_grad(map_loss))

        def step(arg,opt_state):
            loss_value,grads = jit_map_loss(arg) 
            updates, opt_state = optimizer.update(grads, opt_state,arg)
            arg = optax.apply_updates(arg,updates)
            return arg, opt_state, loss_value
        
        for i in range(niter): 
            arg,opt_state,loss_value = step(arg,opt_state)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))
        return arg
    return f 

####################################################################################################################################
####################################################################################################################################
### Registration ###
####################################################################################################################################
####################################################################################################################################


def RegistrationOptimizer(loss:callable,niter=100,optimizer = optax.adam(learning_rate=0.1),verbose=True):
    def f(p0,q0,q0_mask,q1,q1_mask):

        p=p0
        opt_state = optimizer.init(p)
        jit_loss = jit(value_and_grad(loss))

        def step(p,q0,q0_mask,q1,q1_mask,opt_state):
            loss_value,grad = jit_loss(p,q0,q0_mask,q1,q1_mask) 
            updates, opt_state = optimizer.update(grad, opt_state,p)
            p = optax.apply_updates(p,updates)
            return p, opt_state, loss_value
        
        for i in range(niter): 
            p,opt_state,loss_value = step(p,q0,q0_mask,q1,q1_mask,opt_state)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))
        return p
    return f 

def BatchOneToManyRegistrationOptimizerOld(loss:callable,niter=100,optimizer = optax.adam(learning_rate=0.1),verbose=True):
    def f(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask):

        n_batches,batch_size = batched_p0.shape[0],batched_p0.shape[1]
        batched_p=batched_p0
        opt_state = optimizer.init(batched_p)
        vmap_loss = vmap(loss, (0,None,None,0,0),0)
        def rloss(p0,q0,q0_mask,q1,q1_mask): 
            return jnp.sum(vmap_loss(p0,q0,q0_mask,q1,q1_mask))
        jit_loss = jit(value_and_grad(rloss))

        def step(batched_p,q,q_mask,batched_q1,batched_q1_mask,opt_state):

            p_grad = np.zeros_like(batched_p)
            loss_value = 0

            for i,(p,q1,q1_mask) in enumerate(zip(batched_p,batched_q1,batched_q1_mask)):
                t_loss_value,grads = jit_loss(p,q,q_mask,q1,q1_mask)
                p_grad[i] = grads
                loss_value += t_loss_value

           
            loss_value /= (n_batches*batch_size)
            updates, opt_state = optimizer.update(p_grad, opt_state,batched_p)
            batched_p = optax.apply_updates(batched_p,updates)

            return batched_p, opt_state, loss_value
        
        for i in range(niter): 
            batched_p,opt_state,loss_value = step(batched_p,q0,q0_mask,batched_q1,batched_q1_mask,opt_state)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))
        return batched_p
    return f 

def BatchOneToManyRegistrationOptimizer(loss:callable,niter=100,optimizer = optax.adam(learning_rate=0.1),verbose=True):
    def f(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask):

        n_batches,batch_size = batched_p0.shape[0],batched_p0.shape[1]
        batched_p=batched_p0
        opt_state = optimizer.init(batched_p)
        vmap_loss = vmap(loss, (0,None,None,0,0),0)
        def rloss(p0,q0,q0_mask,q1,q1_mask): 
            return jnp.sum(vmap_loss(p0,q0,q0_mask,q1,q1_mask))
        jit_loss = jit(value_and_grad(rloss))

        def step(batched_p,q,q_mask,batched_q1,batched_q1_mask,opt_state):

            p_grad = np.zeros_like(batched_p)
            loss_value = 0

            for i,(p,q1,q1_mask) in enumerate(zip(batched_p,batched_q1,batched_q1_mask)):
                t_loss_value,grads = jit_loss(p,q,q_mask,q1,q1_mask)
                p_grad[i] = grads
                loss_value += t_loss_value

           
            loss_value /= (n_batches*batch_size)
            updates, opt_state = optimizer.update(p_grad, opt_state,batched_p)
            batched_p = optax.apply_updates(batched_p,updates)

            return batched_p, opt_state, loss_value
        
        for i in range(niter): 
            batched_p,opt_state,loss_value = step(batched_p,q0,q0_mask,batched_q1,batched_q1_mask,opt_state)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))
            if i==0: 
                starting_loss = loss_value
        ending_loss = loss_value
        return batched_p, starting_loss, ending_loss
    return f 


####################################################################################################################################
####################################################################################################################################
### Barycenter ###
####################################################################################################################################
####################################################################################################################################

def BatchBarycenterTimeOptimizer(loss:callable,niter=100,optimizer = optax.adam(learning_rate=0.1),verbose=True):
    def f(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask):
        
        n_batches,batch_size = batched_p0.shape[0],batched_p0.shape[1]
        t_q0,s_q0 = q0[:,:1],q0[:,1:]
        batched_p,t_q = batched_p0, t_q0
        opt_state = optimizer.init((batched_p,t_q))
        jit_loss = jit(value_and_grad(loss,argnums=(0,1)))

        def step(batched_p,t_q,s_q,q_mask,batched_q1,batched_q1_mask,opt_state):

            p_grad = np.zeros_like(batched_p)
            loss_value = 0
            t_q_grad = np.zeros_like(t_q)

            for i,(p,q1,q1_mask) in enumerate(zip(batched_p,batched_q1,batched_q1_mask)):
                t_loss_value,grads = jit_loss(p,t_q,s_q,q_mask,q1,q1_mask)
                p_grad[i] = grads[0]
                t_q_grad += grads[1]/(n_batches*batch_size)
                loss_value += t_loss_value

            loss_value /= (n_batches*batch_size)
            updates, opt_state = optimizer.update((p_grad,t_q_grad), opt_state,(batched_p,t_q))
            batched_p,t_q = optax.apply_updates((batched_p,t_q),updates)
            return batched_p, t_q, opt_state, loss_value
        

        for i in range(niter): 
            batched_p,t_q,opt_state,loss_value = step(batched_p,t_q,s_q0,q0_mask,batched_q1,batched_q1_mask,opt_state)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))

        q = jnp.concatenate((t_q,s_q0),axis=-1)
        
        return batched_p,q
    return f 


def BatchBarycenterOptimizer(loss:callable,niter=100,optimizer = optax.adam(learning_rate=0.1),verbose=True):
    def f(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask):
        
        n_batches,batch_size = batched_p0.shape[0],batched_p0.shape[1]
        batched_p,q = batched_p0, q0
        opt_state = optimizer.init((batched_p,q))
        jit_loss = jit(value_and_grad(loss,argnums=(0,1)))

        def step(batched_p,q,q_mask,batched_q1,batched_q1_mask,opt_state):

            p_grad = np.zeros_like(batched_p)
            loss_value = 0
            q_grad = np.zeros_like(q)

            for i,(p,q1,q1_mask) in enumerate(zip(batched_p,batched_q1,batched_q1_mask)):
                t_loss_value,grads = jit_loss(p,q,q_mask,q1,q1_mask)
                p_grad[i] = grads[0]
                q_grad += grads[1]/(n_batches*batch_size)
                loss_value += t_loss_value

            loss_value /= (n_batches*batch_size)
            updates, opt_state = optimizer.update((p_grad,q_grad), opt_state,(batched_p,q))
            batched_p,q = optax.apply_updates((batched_p,q),updates)

            return batched_p, q, opt_state, loss_value
        

        for i in range(niter): 
            batched_p,q,opt_state,loss_value = step(batched_p,q,q0_mask,batched_q1,batched_q1_mask,opt_state)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))
        
        return batched_p,q
    return f 


def BatchIteratedBarycenterOptimizer(shoot:callable,loss:callable,niter=200, update_interval =100, optimizer = optax.adam(learning_rate=0.1),verbose=True):
    def f(batched_p0,q0,q0_mask,batched_q1,batched_q1_mask):
        
        n_batches,batch_size = batched_p0.shape[0],batched_p0.shape[1]
        batched_p,q = batched_p0, q0
        opt_state = optimizer.init((batched_p))
        jit_loss = jit(value_and_grad(loss))

        def step(batched_p,q,q_mask,batched_q1,batched_q1_mask,opt_state):

            p_grad = np.zeros_like(batched_p)
            loss_value = 0

            for i,(p,q1,q1_mask) in enumerate(zip(batched_p,batched_q1,batched_q1_mask)):
                t_loss_value,grads = jit_loss(p,q,q_mask,q1,q1_mask)
                p_grad[i] = grads[0]
                loss_value += t_loss_value

            loss_value /= (n_batches*batch_size)
            updates, opt_state = optimizer.update(p_grad, opt_state,batched_p)
            batched_p = optax.apply_updates(batched_p,updates)

            return batched_p, opt_state, loss_value
        

        for i in range(niter): 
            batched_p,opt_state,loss_value = step(batched_p,q,q0_mask,batched_q1,batched_q1_mask,opt_state)
            if verbose: 
                if (i==0)+((i+1)%10 == 0):
                    print("iteration: ",i+1,"/",niter, " -- loss: ", "{:0.2f}".format(loss_value))
                if ((i+1)!=niter)*(i!=0)*((i+1)%update_interval == 0):
                    print("...Updating barycenter...")
                    pbar = np.mean(batched_p,axis=(0,1))
                    _,q = shoot(pbar,q,q0_mask)
                    batched_p = batched_p - pbar[None,:,:]        
        return batched_p,q
    return f 







from .kernel import *
from .barycenter import *
from .lddmm import *
from .loss import *
from .classifier import *
from .statistic import *
from .plotting import *
from .utils import *
from .TS_PCA_class import *
from .dataset import *




import jax.numpy as jnp
import numpy as np
import jax

# à voir avec thibaut

def _compCLN(x:jnp.ndarray): 
    x0, x1 = x[:-1,:], x[1:,:]
    p = .5*(x0+x1)
    v = (x1-x0)
    m = jnp.sum(v**2,axis=1).reshape(-1,1)
    m = jnp.sqrt(jnp.where(m>0,m,0))
    v = v*jnp.where(m==0.,0.,1./m)
    return m,p,v


def _compELT(x:jnp.ndarray): 
    t_x0, s_x0, t_x1, s_x1 = x[:-1,:1], x[:-1,1:], x[1:,:1],x[1:,1:]
    t_p, s_p = .5*(t_x0+t_x1), .5*(s_x0+s_x1)
    t_v = t_x1 - t_x0
    s_v = (s_x1-s_x0)*jnp.where(t_v==0.,0.,1/t_v)
    return t_p,s_p,t_v,s_v


def _MyLoss(K:callable): 
    def loss(s_x,mask_s_x,t_x,mask_t_x):
        t_t_p,t_s_p,t_t_v,t_s_v = _compELT(t_x)
        s_t_p,s_s_p,s_t_v,s_s_v = _compELT(s_x)
        c0 = jnp.sum(t_t_v * K(t_t_p,t_s_p,t_s_v,mask_t_x[1:,:],t_t_p,t_s_p,t_s_v,mask_t_x[1:,:],t_t_v))
        c1 = jnp.sum(s_t_v * K(s_t_p,s_s_p,s_s_v,mask_s_x[1:,:],t_t_p,t_s_p,t_s_v,mask_t_x[1:,:],t_t_v))
        c2 = jnp.sum(s_t_v * K(s_t_p,s_s_p,s_s_v,mask_s_x[1:,:],s_t_p,s_s_p,s_s_v,mask_s_x[1:,:],s_t_v))
        return c0 -2*c1 +c2
    return loss



def VarifoldLoss(K:callable): 
    """Generate Varifold loss function related to the given varifold kernel

    Parameters
    ----------
        K : kernel function (X,mask_X,Y,mask_Y,b)-> array of the shape of b
            :math:`K(X,Y)b` where X and Y are array of size (n_samples,d+1), :math:`K(X,Y)` is the kernel matrix :math:`(k(x_i,y_j))`
            and b is an array of shape (n_samples,d) with d the dimension of the problem

    Returns
    -------
        Varifoldloss : (s_x,mask_s_x,t_x,mask_t_x)->float
            where s_x,t_x are time series graph (array of size (n_samples,d+1))
            and mask_s_x,mask_t_x are time series graph mask (array of size (n_samples,1))
    """
    def loss(s_x,mask_s_x,t_x,mask_t_x): 
        t_m,t_p,t_v = _compCLN(t_x)
        s_m,s_p,s_v = _compCLN(s_x)
        c0 = jnp.sum(t_m * K(t_p,t_p,t_v,t_v,mask_t_x[1:,:],mask_t_x[1:,:],t_m))
        c1 = jnp.sum(s_m * K(s_p,t_p,s_v,t_v,mask_s_x[1:,:],mask_t_x[1:,:],t_m))
        c2 = jnp.sum(s_m * K(s_p,s_p,s_v,s_v,mask_s_x[1:,:],mask_s_x[1:,:],s_m))
        return c0 -2*c1 + c2
    return loss

def SumVarifoldLoss(K_lst:callable): 
    """Generate a sum of Varifold loss function related to the given varifold kernels
    
    Parameters
    ----------
        K_lst : list of kernel functions (X,mask_X,Y,mask_Y,b)-> array of the shape of b
            :math:`K(X,Y)b` where X and Y are array of size (n_samples,d+1), :math:`K(X,Y)` is the kernel matrix :math:`(k(x_i,y_j))`
            and b is an array of shape (n_samples,d) with d the dimension of the problem

    Returns
    -------
        Sum of Varifoldloss : (s_x,mask_s_x,t_x,mask_t_x)->float
            where s_x,t_x are time series graph (array of size (n_samples,d+1))
            and mask_s_x,mask_t_x are time series graph mask (array of size (n_samples,1))
    """
    def loss(s_x,mask_s_x,t_x,mask_t_x): 
        t_m,t_p,t_v = _compCLN(t_x)
        s_m,s_p,s_v = _compCLN(s_x)
        score = 0
        for K in K_lst:
            c0 = jnp.sum(t_m * K(t_p,t_p,t_v,t_v,mask_t_x[1:,:],mask_t_x[1:,:],t_m))
            c1 = jnp.sum(s_m * K(s_p,t_p,s_v,t_v,mask_s_x[1:,:],mask_t_x[1:,:],t_m))
            c2 = jnp.sum(s_m * K(s_p,s_p,s_v,s_v,mask_s_x[1:,:],mask_s_x[1:,:],s_m))
            score +=  c0 -2*c1 + c2
        return score
    return loss

def _MMD(K:callable): 
    def loss(s_x,mask_s_x,t_x,mask_t_x): 
        ns = s_x.shape[0]
        s_ones = np.ones((ns,1))/np.sum(mask_s_x).astype(np.float32)
        t_ones = np.ones((ns,1))/np.sum(mask_t_x).astype(np.float32)

        c0 = jnp.sum(t_ones*K(t_x,mask_t_x,t_x,mask_t_x,t_ones))
        c1 = jnp.sum(s_ones*K(s_x,mask_s_x,t_x,mask_t_x,t_ones))
        c2 = jnp.sum(s_ones*K(s_x,mask_s_x,s_x,mask_s_x,s_ones))
        return c0 -2*c1 + c2
    return loss

def _WeightedMMD(K:callable): 
    def loss(s_x,mask_s_x,t_x,mask_t_x): 
        t_m,t_p,t_v = _compCLN(t_x)
        s_m,s_p,s_v = _compCLN(s_x)
        c0 = jnp.sum(t_m * K(t_p,mask_t_x[1:,:],t_p,mask_t_x[1:,:],t_m))
        c1 = jnp.sum(s_m * K(s_p,mask_s_x[1:,:],t_p,mask_t_x[1:,:],t_m))
        c2 = jnp.sum(s_m * K(s_p,mask_s_x[1:,:],s_p,mask_s_x[1:,:],s_m))
        return c0 -2*c1 + c2
    return loss

def _MomentaLoss(K:callable,q0:jnp.ndarray,q0_mask): 
    def loss(s_p,t_p): 
        c0 = jnp.sum(s_p * K(q0,q0_mask,q0,q0_mask,s_p))
        c1 = jnp.sum(t_p * K(q0,q0_mask,q0,q0_mask,s_p))
        c2 = jnp.sum(t_p * K(q0,q0_mask,q0,q0_mask,t_p))
        return c0 -2*c1 + c2
    return loss


import numpy as np
import jax.numpy as jnp
from jax import vmap 


def time_shape_embedding(x:np.ndarray,sampfreq=1,max_length=None,dtype = jnp.float32): 
    """time series to time series'graph 

    Parameters
    ----------
        x : (array) 
            signal (T,d) or ()
        sampfreq : (int, optional)
             sampling frequency. Defaults to 1. Important to set when it's not 1 for hyperparameter tuning
        max_length : (int, optional): 
            maximum lenght desired. Defaults to None.
        dtype : (_type_, optional):
             float32 or 64 depending on the memory. Defaults to jnp.float32.

    Returns
    -------
    pair
        The graph of the signal x : array of shape (T,d+1)
        Its mask  : array of shape (T,1)
    """
    n_s = x.shape[0]
    time = np.arange(n_s,dtype=float).reshape(-1,1)/sampfreq
    t_x = np.hstack((time,x),dtype=dtype)
    if max_length is None:
        return t_x
    else:
        t_x = np.pad(t_x,((0,max_length-n_s),(0,0))).astype(dtype)
        mask = np.full_like(t_x[:,:1],True,dtype=bool)
        mask[n_s:,:] = False
        return t_x, mask

def from_timeseries_to_dataset(X:list,sampfreq=1,dtype=jnp.float32): 
    """
    Convert a list of time series arrays into a fixed-shape dataset suitable for TS-PCA.

    Each time series is transformed into a time-augmented array of shape 
    :math:`(n\_samples, d+1)` where the first column contains time points (with 
    sampling frequency `sampfreq`) and the remaining columns contain the original 
    time series values. The function also returns a mask indicating valid samples 
    for each time series.

    Parameters
    ----------
        X : list of arrays
            List of time series arrays. Each element should have shape `(n_samples, d)` 
            or `(n_samples,)` for univariate series.

        sampfreq : float, optional
            Sampling frequency used to generate the time points. Default is 1.

        dtype : jnp.dtype, optional
            Data type of the returned arrays. Default is `jnp.float32`.

    Returns
    -------
        dataset : jnp.array of shape (n_time_series, n_samples_max, d+1)
            Padded dataset of time series, where `n_samples_max` is the length of the 
            longest time series. The first column contains time points, and the 
            remaining columns contain time series values.

        mask : jnp.array of shape (n_time_series, n_samples_max, 1)
            Boolean mask indicating valid samples for each time series. True indicates 
            a valid entry, False indicates padding.

    Notes
    -----
        - The function pads shorter time series to the length of the longest series in `X`.
        - Internally, it calls `time_shape_embedding` to embed each time series with time 
        information and generate the corresponding mask.
        - Suitable for use as input to `TS-PCA` methods that require a fixed-size dataset 
        and mask.
    """
    lengths = np.array([x.shape[0] for x in X])
    max_length = np.max(lengths)
    mask_lst = []
    ts_lst = []
    for ts in X: 
        t_ts, t_mask = time_shape_embedding(ts,sampfreq,max_length,dtype=dtype)
        ts_lst.append(t_ts)
        mask_lst.append(t_mask)
    return jnp.array(ts_lst,dtype=dtype), jnp.array(mask_lst,dtype=jnp.bool_)
    
def batch_dataset(dataset,batch_size,masks=None,dtype=jnp.float32): 
    """
    Change the shape to apply a batch function

    Parameters
    ----------
        dataset : array of shape (n_ts,n_s,n_d)
            dataset of time series graph
        batch_size : (int)
            we should have n_ts // batch_size=int
        masks : array (n_ts,n_s,1) 
            Mask related to the dataset
       
    Raises
    ------
        ValueError: dataset size is not a multiple of batch_size

    Returns
    -------
    pair
        reshaped dataset : array of shape (n_batches,batch_size,n_s,n_d)
         
        reshaped mask : array of shape (n_batches,batch_size,n_s,1) 
    """
    if dataset.shape[0]%batch_size!=0: 
        raise ValueError("dataset size is not a multiple of batch_size")
    n_ts,n_s,n_d = dataset.shape
    n_batches = n_ts//batch_size
    if masks is None: 
        return dataset.reshape(n_batches,batch_size,n_s,n_d).astype(dtype)
    else: 
        return dataset.reshape(n_batches,batch_size,n_s,n_d).astype(dtype), masks.reshape(n_batches,batch_size,n_s,1)
    
def unbatch_dataset(batched_dataset,batched_masks=None,dtype=jnp.float32): 
    """
    Change the shape to remove batch dim

    Parameters
    ----------
        batched_dataset : array of shape (n_batches,batch_size,n_s,n_d)
            batch dataset of time series graph
    
        batched_masks : array (n_batches,batch_size,n_s,1)  
            Mask related to the dataset
       

    Returns
    -------
    pair
        dataset : array of shape (n,n_s,n_d)
         
        mask : array of shape (n,n_s,1) 
    """
    _,_,n_s,n_d = batched_dataset.shape
    if batched_masks is None: 
        return batched_dataset.reshape(-1,n_s,n_d).astype(dtype)
    else: 
        return batched_dataset.reshape(-1,n_s,n_d).astype(dtype), batched_masks.reshape(-1,n_s,1)
    
